/* This program plays the game of GOBANG. In this game the player
   and the computer place stones on the board and the one that
   first makes a row of 5 wins. The computer plays in a very
   simple, and, hence, fast way. Up to level 5 no backtracking or such
   is performed. The program uses nothing but simple
   position evaluation. Still he is quite strong. The interaction
   to this game was created using the Forms library.

   Written by Mark Overmars

   Version 4.0
   Date 26 Nov 1990
*/

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <gl/gl.h>
#include "forms.h"

/* The stones */
#define NOONE 0
#define WST 10
#define BST 100
#define BORDER 1000

/* The field */
#define FSIZE   19
#define MAXMOVE FSIZE * FSIZE


int field[FSIZE+8][FSIZE+8];            /* The field */
int possible[FSIZE+8][FSIZE+8];         /* Possible places to move */
int numb;                               /* Number of moves sofar */
int level;                              /* Playing level of computer */
int gameover;                           /* Whether game is over */
int gamenumb,mescore,youscore;          /* Number of the game and score sofar */
int lastwx,lastwy,lastbx,lastby;        /* Last position played */
int winx[5],winy[5];			/* The winning positions */

/* The test performed for the different levels. */
int tests[8][10]={{3,4,11,12,0,0,0,0,0,0},
                  {4,4,6,11,12,0,0,0,0,0},
                  {6,4,5,6,8,11,12,0,0,0},
                  {7,4,5,6,7,8,11,12,0,0},
                  {9,4,5,6,7,8,9,10,11,12},
                  {9,4,5,6,7,8,9,10,11,12},
                  {9,4,5,6,7,8,9,10,11,12},
                  {9,4,5,6,7,8,9,10,11,12} };

/**** SOME HELP PROCEDURES ****/

int empty(int a,int b)      /* Checks whether position is empty */
{  return (field[a][b] == NOONE); }

int hor(int a,int b,int n)  /* calculates the sum of a horizontal part */
{ int i, k=0; for (i=0;i<n;i++) k += field[a+i][b]; return k;}

int vert(int a,int b,int n) /* calculates the sum of a vertical part */
{ int i, k=0; for (i=0;i<n;i++) k += field[a][b+i]; return k;}

int up(int a,int b,int n)   /* calculates the sum of a upwards diagonal part */
{ int i, k=0; for (i=0;i<n;i++) k += field[a+i][b+i]; return k;}

int down(int a,int b,int n) /* calculates sum of a downwards diagonal part */
{ int i, k=0; for (i=0;i<n;i++) k += field[a+i][b-i]; return k;}

int win_test(int x,int y,int stone)
/* Checks whether position x,y is winning */
{
  int i,j,a,b,c;
  for (i = -4; i<=0; i++)
  {
    a=x+i; b=y+i; c=y-i;
    if (hor(a,y,5) == 5*stone)
      {for (j=0;j<5;j++) {winx[j]=a+j; winy[j]=y;}  return 1;}
    if (up(a,b,5) == 5*stone)
      {for (j=0;j<5;j++) {winx[j]=a+j; winy[j]=b+j;}  return 1;}
    if (down(a,c,5) == 5*stone)
      {for (j=0;j<5;j++) {winx[j]=a+j; winy[j]=c-j;}  return 1;}
    if (vert(x,b,5) == 5*stone)
      {for (j=0;j<5;j++) {winx[j]=x; winy[j]=b+j;}  return 1;}
  }
  return 0;
}

void check_winner(int x,int y,int stone)
/* Checks whether this move causes a winner. */
{
  if (win_test(x,y,stone))
  {
    gameover = TRUE;
    if (stone == WST) youscore++; else mescore++;
  }
}

/***** ROUTINES FOR EVALUATING POSITIONS ****/

int checkKin5(int x,int y,int stone,int K)
/* Checks in how many rows of lenght K within 5 the stone at x,y is involved.*/
{
  int i, a, b, c, n=0;
  for (i = -4;i<=0;i++)
  {
    a=x+i; b=y+i; c=y-i;
    if (hor(a,y,5)  == K*stone) n++;
    if (up(a,b,5)   == K*stone) n++;
    if (down(a,c,5) == K*stone) n++;
    if (vert(x,b,5) == K*stone) n++;
  }
  return n;
}

int check_Kin4_(int x,int y,int stone,int K)
/* Checks in how many rows of K in empty 4 empty x,y is involved. */
{
  int a, b, c, i, n1=0, n2=0, n3=0, n4=0;
  for (i = -3; i<=0; i++)
  {
    a=x+i; b=y+i; c=y-i;
    if (hor(a,y,4)  == K*stone && empty(a-1,y)   && empty(a+4,y) )   n1 = 1;
    if (up(a,b,4)   == K*stone && empty(a-1,b-1) && empty(a+4,b+4) ) n2 = 1;
    if (down(a,c,4) == K*stone && empty(a-1,c+1) && empty(a+4,c-4) ) n3 = 1;
    if (vert(x,b,4) == K*stone && empty(x,b-1)   && empty(x,b+4) )   n4 = 1;
  }
  return (n1+n2+n3+n4);
}

int check_3in3_(int x,int y,int stone)
/* Checks in how many rows of K in empty 3 empty x,y is involved. */
{
  int a, b, c,i, n1=0, n2=0, n3=0, n4=0;
  for (i = -2; i<=0; i++)
  {
    a=x+i; b=y+i; c=y-i;
    if (hor(a,y,3)  == 3*stone && empty(a-1,y)   && empty(a+3,y) )   n1 = 1;
    if (up(a,b,3)   == 3*stone && empty(a-1,b-1) && empty(a+3,b+3) ) n2 = 1;
    if (down(a,c,3) == 3*stone && empty(a-1,c+1) && empty(a+3,c-3) ) n3 = 1;
    if (vert(x,b,3) == 3*stone && empty(x,b-1)   && empty(x,b+3) )   n4 = 1;
  }
  return (n1+n2+n3+n4);
}

int checkthrees(int x,int y,int stone)
/* Checks in how many open rows of lenght 3 x,y is involved. */
{
  int a, b, c, i, n1=0, n2=0, n3=0, n4=0;
  for (i = -4; i<=0; i++)
  {
    a=x+i; b=y+i; c=y-i;
    if (hor(a,y,5) == 4*stone ||
      (hor(a,y,4)  == 3*stone && empty(a-1,y)   && empty(a+4,y) ))   n1 = 1;
    if (up(a,b,5) == 4*stone ||
      (up(a,b,4)   == 3*stone && empty(a-1,b-1) && empty(a+4,b+4) )) n2 = 1;
    if (down(a,c,5) == 4*stone ||
      (down(a,c,4) == 3*stone && empty(a-1,c+1) && empty(a+4,c-4) )) n3 = 1;
    if (vert(x,b,5) == 4*stone ||
      (vert(x,b,4) == 3*stone && empty(x,b-1)   && empty(x,b+4) ))   n4 = 1;
  }
  return (n1+n2+n3+n4);
}

/**** THE BACKTRACKING   ****/
int xpos[361],ypos[361],l,newl;	/* list of possible moves */
int steps,maxsteps;		/* steps allowed in recursion */

void simple_place(int x,int y,int stone)
{
  field[x][y] = stone;
  possible[x][y] = (stone == NOONE);
}

int track_black(int);

int track_white(int d)
{
  int i,x,y,tt;
  if (steps++ > maxsteps) return -1;
  for (i=0; i<l; i++)
  {
    x = xpos[i];  y = ypos[i];
    if (possible[x][y])
    {
      simple_place(x,y,BST);
      if (checkKin5(x,y,BST,5))
      {
        simple_place(x,y,WST);
        if (checkKin5(x,y,WST,4))
          { simple_place(x,y,NOONE); return 0;}
        tt = track_black(d);
        simple_place(x,y,NOONE);
        return tt;
      }
      simple_place(x,y,NOONE);
    }
  }
  return 0;
}

int track_black(int d)
{
  int i, x, y, max = 0, tt;
  if (steps++ > maxsteps) return -1;
  for (i=0; i<l; i++)
  {
    x = xpos[i]; y = ypos[i];
    if (possible[x][y])
    {
      simple_place(x,y,BST);
      if (checkKin5(x,y,BST,4))
      {
        if (checkthrees(x,y,BST) > 1)
          {simple_place(x,y,NOONE); return 10+d;}
        else if (d>1)
        {
          tt = track_white(d-1);
          simple_place(x,y,NOONE);
          if (tt == -1) return -1;
          if (tt > 10) return tt;
          if (tt > max) max = tt;
        }
      }
      simple_place(x,y,NOONE);
    }
  }
  for (i=0; i<l; i++)
  {
    x = xpos[i]; y = ypos[i];
    if (possible[x][y])
    {
      simple_place(x,y,BST);
      if (checkthrees(x,y,BST)>1)
        {simple_place(x,y,NOONE); return d;}
      else
        simple_place(x,y,NOONE);
    }
  }
  return max;
}

int track_test(int x,int y,int d)
/* does a test with backtracking */
{
  int tt = 0;
  if (! possible[x][y]) return 0;
  simple_place(x,y,BST);
  if (checkKin5(x,y,BST,4))
  {
    steps = 0;
    maxsteps = 30*d*d*d;
    tt = track_white(d);
  }
  simple_place(x,y,NOONE);
  return tt;
}

/**** FINDING GOOD MOVES ****/

int perform_test(int testnum,int x,int y,int stone)
/* performs test testnum on position x,y for stone.*/
{
  int temp,player,other;
  if (stone==BST) {player=BST;other=WST;} else {player=WST;other=BST;}
  if (testnum % 2 ==1) field[x][y]=player; else field[x][y]=other;
  switch (testnum)
  {
    case 1: temp=checkKin5(x,y,player,5);break;
    case 2: temp=checkKin5(x,y,other,5);break;
    case 3: temp=checkKin5(x,y,player,4);
             if (temp==1 && checkthrees(x,y,player)<=1) temp=0;break;
    case 4: temp=checkKin5(x,y,other,4);
             if (temp==1 && checkthrees(x,y,other)<=1) temp=0;break;
    case 5: temp=checkthrees(x,y,player);
             if (numb<6 || temp<=1) temp=0;break;
    case 6: temp=checkthrees(x,y,other);
             if (numb<5 || temp<=1) temp=0;break;
    case 7: temp=checkthrees(x,y,player);
             if (numb<6) temp=0;break;
    case 8: temp=checkthrees(x,y,other);
             if (numb<5) temp=0;break;
    case 9: temp=check_3in3_(x,y,player);
             if (numb<6) temp=0;break;
    case 10: temp=check_3in3_(x,y,other);
             if (numb<5) temp=0;break;
    case 11: temp=check_Kin4_(x,y,player,2);break;
    case 12: temp=check_Kin4_(x,y,other,2);break;
  }
  field[x][y]=NOONE;
  return(temp);
}

void find_move(int stone,int *xx,int *yy)
/* Returns the best move for stone. */
{   
  int count,temp,i,j;
  /* make a list of possible moves. */
  l = 0;
  for (i = 4; i < FSIZE+4; i++) for (j = 4; j < FSIZE+4; j++)
      if (possible[i][j]) {xpos[l] = i; ypos[l] = j; l++;}
  /* test the list of possible moves, every time refining the list. */
  /* simple tests */
  for (i = 1; i <= 3; i++)
  {
    count = 0; newl = 0;
    for (j=0; j<l; j++)
    {
      temp=perform_test(i,xpos[j],ypos[j],stone);
      if (temp >  count) {count =  temp; newl = 0;}
      if (temp == count) {xpos[newl] = xpos[j]; ypos[newl] = ypos[j]; newl++;}
    }
    l = newl;
  }
  /* For high levels do backtracking */
  if (level >= 6 && stone == BST)
  {
    count = 0;
    for (j=0; j<l; j++)
    {
      temp = track_test(xpos[j],ypos[j],level-3);
      if (temp > count)
        {count = temp; *xx = xpos[j]; *yy = ypos[j];}
      if (count > 10) return;
    }
    temp = 1;
    for (j=0; j<l; j++)
      if (perform_test(4,xpos[j],ypos[j],stone) ||
		 perform_test(5,xpos[j],ypos[j],stone))
        temp = 0;
    if (temp && count>0) return;
  }
  /* Do other tests */
  for (i = 1; i <= tests[level-1][0]; i++)
  {
    count = 0; newl = 0;
    for (j=0; j<l; j++)
    {
      temp=perform_test(tests[level-1][i],xpos[j],ypos[j],stone);
      if (temp >  count) {count =  temp; newl = 0;}
      if (temp == count) {xpos[newl] = xpos[j]; ypos[newl] = ypos[j]; newl++;}
    }
    l = newl;
  }
  /* Take a possible move*/
  if (l==0) { xpos[0]=13; ypos[0]=13; l=1;};
  i= rand() % l;
  *xx = xpos[i]; *yy = ypos[i];
}

void place(int x,int y,int stone)
/* places the stone on position x,y and updates the possible moves. */
{
  int i,j;
  field[x][y] = stone;
  numb++;
  possible[x][y] = (stone == NOONE);
  for (i=x-2; i<=x+2; i++) for (j=y-2; j<=y+2; j++)
      if (field[i][j] == NOONE) possible[i][j] = TRUE;
  if (stone==BST) {lastbx=x;lastby=y;} else {lastwx=x;lastwy=y;}
  check_winner(x,y,stone);
}

void init_play()
/* Initializes the playing of games */
{
  gamenumb = 0;
  mescore = 0;
  youscore = 0;
  level = 4;
  srand( (int) (time(0) % 10000));
}

void init_game()
/* Initializes a single game */
{
  int i,j;
  for (i=0;i<FSIZE+8;i++) for (j=0;j<FSIZE+8;j++) field[i][j] = BORDER;
  for (i=4;i<FSIZE+4;i++) for (j=4;j<FSIZE+4;j++) field[i][j] = NOONE;
  for (i=0;i<FSIZE+8;i++) for (j=0;j<FSIZE+8;j++) possible[i][j] = FALSE;
  numb = 0;
  lastwx = -1;
  lastbx = -1;
  gameover = FALSE;
  gamenumb++;
  if (gamenumb%2 == 0) place(13,13,BST);
}


void make_move(int color)
/* Makes the best move for the color */
{
  int xx,yy;
  if (! gameover)
  {
    find_move(color,&xx,&yy);
    place(xx,yy,color);
  }
}

void undo_move()
/* Undoes the last move */
{
  if (!gameover && lastwx!= -1)
  {
    field[lastwx][lastwy]=NOONE;
    field[lastbx][lastby]=NOONE;
    possible[lastwx][lastwy]=TRUE;
    possible[lastbx][lastby]=TRUE;
    lastwx = -1;  lastbx = -1;
  }
}

/*--------------- THE FORM -------------------*/

#define FX      25.0
#define FY      25.0
#define FW      25.0
#define FH      25.0

#define WW      2.0 * FX + FW * FSIZE
#define HH      2.0 * FY + FH * FSIZE 

FL_FORM *form;
FL_OBJECT *level1, *level2, *level3, *level4, *level5, *level6,
       *level7, *fieldobj, *newgame, *endgame, *undomove, *me, *you,
       *help;

FL_FORM *helpform;
FL_OBJECT *helpok;

void draw_field()
/* redraws the field */
{
int i,j;
  fl_color(7);
  rectf(FX,FY,FX+FW*FSIZE,FY+FH*FSIZE);
  fl_color(0);
  for (i=0; i<=FSIZE;i++)
  {
    move2(FX+FW*i,FY); draw2(FX+FW*i,FY+FH*FSIZE);
    move2(FX,FY+FH*i); draw2(FX+FW*FSIZE,FY+FH*i);
  }
  for (i=0; i<FSIZE;i++)
    for (j=0; j<FSIZE;j++)
    {
      switch (field[i+4][j+4])
      {
        case WST: fl_color(1); break;
        case BST: fl_color(4); break;
      }
      if (field[i+4][j+4] != NOONE)
        circf(FX+FW*i+0.5*FW,FY+FH*j+0.5*FH,0.4*FW);
    }
  if (gameover)
  {
    fl_color(7);
    for (i=0; i<5; i++)
      circf(FX+FW*(winx[i]-4)+0.5*FW,FY+FH*(winy[i]-4)+0.5*FH,0.2*FW);
  }
  else if (lastbx != -1)
  {
    fl_color(6);
    circf(FX+FW*(lastbx-4)+0.5*FW,FY+FH*(lastby-4)+0.5*FH,0.2*FW);
  }
}

int handle_field(FL_OBJECT *obj, int event, float mx, float my, char key)
/* The handle routine for the free object that controls
   the field. */
{
  int xx,yy;
  char str[10];
  switch (event)
  {
    case FL_DRAW:
      draw_field();
      return 0;
    case FL_PUSH:
      xx = (int) ((mx-FX)/FW);
      yy = (int) ((my-FY)/FH);
      if (! gameover && field[xx+4][yy+4] == NOONE)
      {
        place(xx+4,yy+4,WST);
	lastbx = -1;
        fl_redraw_object(fieldobj);
        if (! gameover) make_move(BST);
        fl_redraw_object(fieldobj);
	if (gameover)
	{
	  sprintf(str,"%d",mescore);
	  fl_set_object_label(me,str);
	  sprintf(str,"%d",youscore);
	  fl_set_object_label(you,str);
	}
      }
      return 0;
  }
  return 0;
}

#define P 20.0

void make_helpform()
{
  float pos;
  helpform = fl_bgn_form(FL_UP_BOX,300.0,400.0);
  pos = 360.0;
  fl_add_text(FL_NORMAL_TEXT,20.0,(pos -= P),0.0,0.0,
     "The goal of the game of Gobang is to form");
  fl_add_text(FL_NORMAL_TEXT,20.0,(pos -= P),0.0,0.0,
     "a row of five stones of your color, either");
  fl_add_text(FL_NORMAL_TEXT,20.0,(pos -= P),0.0,0.0,
     "horizontal, vertical, or diagonal. The");
  fl_add_text(FL_NORMAL_TEXT,20.0,(pos -= P),0.0,0.0,
     "computer will try to do the same. Who");
  fl_add_text(FL_NORMAL_TEXT,20.0,(pos -= P),0.0,0.0,
     "finishes first has won the game.");
  fl_add_text(FL_NORMAL_TEXT,20.0,(pos -= P),0.0,0.0,
     "");
  fl_add_text(FL_NORMAL_TEXT,20.0,(pos -= P),0.0,0.0,
     "You can set the level by pressing the");
  fl_add_text(FL_NORMAL_TEXT,20.0,(pos -= P),0.0,0.0,
     "appropriate button. At levels larger than");
  fl_add_text(FL_NORMAL_TEXT,20.0,(pos -= P),0.0,0.0,
     "5 the computer might take a bit longer");
  fl_add_text(FL_NORMAL_TEXT,20.0,(pos -= P),0.0,0.0,
     "to find a move. The last move of the");
  fl_add_text(FL_NORMAL_TEXT,20.0,(pos -= P),0.0,0.0,
     "computer is indicated with a dot.");
  fl_add_text(FL_NORMAL_TEXT,20.0,(pos -= P),0.0,0.0,
     "");
  fl_add_text(FL_NORMAL_TEXT,20.0,(pos -= P),0.0,0.0,
     "Good Luck.");

  helpok = fl_add_button(FL_NORMAL_BUTTON,120.0,pos-80.0,80.0,30.0,"Ready");
}

void makeform()
{
  FL_OBJECT *obj;
  form = fl_bgn_form(FL_UP_BOX,WW+275.0,HH);
    fl_add_box(FL_DOWN_BOX,FX-10.0,FY-10.0,FW*FSIZE+20.0,FH*FSIZE+20.0,"");
    fieldobj = fl_add_free(FL_NORMAL_FREE,FX,FY,FW*FSIZE,FH*FSIZE,"",handle_field);

    obj = fl_add_box(FL_UP_BOX,WW+25.0,HH-95.0,225.0,80.0,"GOBANG");
    fl_set_object_lsize(obj,FL_LARGE_FONT);
    obj = fl_add_text(FL_NORMAL_TEXT,WW+25.0,HH-90.0,225.0,75.0,"Version 4.0   by Mark Overmars");
    fl_set_object_lsize(obj,FL_SMALL_FONT);
    fl_set_object_align(obj,FL_ALIGN_BOTTOM);

    level1= fl_add_lightbutton(FL_RADIO_BUTTON,WW+35.0,HH-150.0,95.0,30.0,"level 1");
    level2= fl_add_lightbutton(FL_RADIO_BUTTON,WW+35.0,HH-185.0,95.0,30.0,"level 2");
    level3= fl_add_lightbutton(FL_RADIO_BUTTON,WW+35.0,HH-220.0,95.0,30.0,"level 3");
    level4= fl_add_lightbutton(FL_RADIO_BUTTON,WW+35.0,HH-255.0,95.0,30.0,"level 4");
    level5= fl_add_lightbutton(FL_RADIO_BUTTON,WW+145.0,HH-150.0,95.0,30.0,"level 5");
    level6= fl_add_lightbutton(FL_RADIO_BUTTON,WW+145.0,HH-185.0,95.0,30.0,"level 6");
    level7= fl_add_lightbutton(FL_RADIO_BUTTON,WW+145.0,HH-220.0,95.0,30.0,"level 7");

    undomove= fl_add_button(FL_NORMAL_BUTTON,WW+35.0,HH-305.0,95.0,30.0,"undo last");
    help= fl_add_button(FL_NORMAL_BUTTON,WW+145.0,HH-305.0,95.0,30.0,"help");
    newgame= fl_add_button(FL_NORMAL_BUTTON,WW+90.0,HH-345.0,95.0,30.0,"new game");
    endgame= fl_add_button(FL_NORMAL_BUTTON,WW+90.0,HH-500.0,95.0,30.0,"exit");

    fl_add_box(FL_UP_BOX,WW+35.0,HH-455.0,205.0,95.0,"");
    obj = fl_add_text(FL_NORMAL_TEXT,WW+35.0,HH-445.0,205.0,80.0,"Score");
    fl_set_object_align(obj,FL_ALIGN_TOP);
    
    me= fl_add_box(FL_DOWN_BOX,WW+140.0,HH-415.0,80.0,25.0,"0");
    fl_set_object_color(me,13,13);
    obj= fl_add_text(FL_NORMAL_TEXT,WW+140,HH-415,0.0,25.0,"Computer  ");
    fl_set_object_align(obj,FL_ALIGN_RIGHT);
    
    you= fl_add_box(FL_DOWN_BOX,WW+140.0,HH-445.0,80.0,25.0,"0");
    fl_set_object_color(you,13,13);
    obj= fl_add_text(FL_NORMAL_TEXT,WW+140,HH-445,0.0,25.0,"Human  ");
    fl_set_object_align(obj,FL_ALIGN_RIGHT);
  fl_end_form();
}

/*--------------- MAIN CONTROL ---------------*/

void main_loop()
/* The main loop (not doing very much) */
{
  FL_OBJECT *obj;
  while (1)
  {
    obj = fl_do_forms();
    if      (obj == level1) level = 1;
    else if (obj == level2) level = 2;
    else if (obj == level3) level = 3;
    else if (obj == level4) level = 4;
    else if (obj == level5) level = 5;
    else if (obj == level6) level = 6;
    else if (obj == level7) level = 7;
    else if (obj == newgame) {init_game(); fl_redraw_object(fieldobj); }
    else if (obj == endgame) return;
    else if (obj == undomove) {undo_move(); fl_redraw_object(fieldobj); }
    else if (obj == help) {fl_show_form(helpform,FL_PLACE_SIZE,FALSE,NULL);}
    else if (obj == helpok) {fl_hide_form(helpform);}
  }
}

main()
{
  init_play();
  init_game();
  makeform();
  make_helpform();
  fl_set_button(level4,1);
  fl_show_form(form,FL_PLACE_SIZE,TRUE,"Gobang");
  main_loop();
  fl_hide_form(form);
}
